import sys
from datetime import datetime
from importlib.metadata import version
from pathlib import Path
from typing import cast

import emoji
from typer import Exit, secho

from sigexport import models

VERSION = version("signal-export")


def dt_from_ts(ts: float) -> datetime:
    return datetime.fromtimestamp(ts / 1000.0)


def parse_datetime(input_str: str) -> datetime:
    last_exception = None
    for fmt in [
        "%Y-%m-%d %H:%M",
        "%Y-%m-%d, %H:%M",
        "%Y-%m-%d %H:%M:%S",
        "%Y-%m-%d, %H:%M:%S",
    ]:
        try:
            return datetime.strptime(input_str, fmt)
        except ValueError as e:
            last_exception = e
    exception = cast(ValueError, last_exception)
    raise (exception)


def version_callback(value: bool) -> None:
    """Get sigexport version."""
    if value:
        print(f"v{VERSION}")
        raise Exit()


def source_location() -> Path:
    """Get OS-dependent source location."""
    home = Path.home()
    paths = {
        "linux": home / ".config/Signal",
        "linux2": home / ".config/Signal",
        "darwin": home / "Library/Application Support/Signal",
        "win32": home / "AppData/Roaming/Signal",
    }
    try:
        source_path = paths[sys.platform]
    except KeyError:
        secho("Please manually enter Signal location using --source.")
        raise Exit(code=1)

    return source_path


def fix_names(contacts: models.Contacts) -> models.Contacts:
    """Convert contact names to filesystem-friendly."""
    fixed_contact_names = set()
    for key, item in contacts.items():
        contact_name = item.number if item.name is None else item.name
        if contacts[key].name is not None:
            contacts[key].name = "".join(
                x for x in emoji.demojize(contact_name) if x.isalnum()
            )
            if contacts[key].name == "":
                contacts[key].name = "unnamed"
            fixed_contact_name = contacts[key].name
            if fixed_contact_name in fixed_contact_names:
                name_differentiating_number = 2
                while (
                    fixed_contact_name + str(name_differentiating_number)
                ) in fixed_contact_names:
                    name_differentiating_number += 1
                fixed_contact_name += str(name_differentiating_number)
                contacts[key].name = fixed_contact_name
            fixed_contact_names.add(fixed_contact_name)

    return contacts



import shutil
from collections.abc import Iterable
from datetime import datetime
from pathlib import Path

from typer import colors

from sigexport import models
from sigexport.logging import log


def copy_attachments(
    src: Path, dest: Path, convos: models.Convos, contacts: models.Contacts
) -> Iterable[tuple[Path, Path]]:
    """Copy attachments and reorganise in destination directory."""
    src_att = Path(src) / "attachments.noindex"
    dest = Path(dest)

    for key, messages in convos.items():
        name = contacts[key].name
        log(f"\tCopying attachments for: {name}")
        # some contact names are None
        if not name:
            name = "None"
        contact_path = dest / name / "media"
        contact_path.mkdir(exist_ok=True, parents=True)
        for msg in messages:
            if msg.attachments:
                attachments = msg.attachments
                date = (
                    datetime.fromtimestamp(msg.get_ts() / 1000)
                    .isoformat(timespec="milliseconds")
                    .replace(":", "-")
                )
                for i, att in enumerate(attachments):
                    try:
                        # Account for no fileName key
                        file_name = (
                            str(att["fileName"]) if "fileName" in att else "None"
                        )
                        # Sometimes the key is there but it is None, needs extension
                        if "." not in file_name:
                            content_type = att["contentType"].split("/")
                            try:
                                ext = content_type[1]
                            except IndexError:
                                ext = content_type[0]
                            file_name += "." + ext
                        att["fileName"] = (
                            f"{date}_{i:02}_{file_name}".replace(" ", "_")
                            .replace("/", "-")
                            .replace(",", "")
                            .replace(":", "-")
                            .replace("|", "-")
                        )
                        # account for erroneous backslash in path
                        att_path = str(att["path"]).replace("\\", "/")
                        yield src_att / att_path, contact_path / att["fileName"]
                    except KeyError:
                        p = att["path"] if "path" in att else ""
                        log(f"\t\tBroken attachment:\t{name}\t{p}")
                    except FileNotFoundError:
                        p = att["path"] if "path" in att else ""
                        log(f"\t\tAttachment not found:\t{name}\t{p}")
            else:
                msg.attachments = []


def merge_attachments(media_new: Path, media_old: Path) -> None:
    """Merge new and old attachments directories."""
    for f in media_old.iterdir():
        if f.is_file():
            try:
                shutil.copy2(f, media_new)
            except shutil.SameFileError:
                log(
                    f"Skipped file {f} as duplicate found in new export directory!",
                    fg=colors.RED,
                )



import os
import shutil
import tempfile
import time
from pathlib import Path

os.environ["TZ"] = "UTC"

time.tzset()

expected_md = """[2022-08-10 19:33:38] Me: Test message  

[2022-08-10 19:33:48] Me: Test image  ![2022-08-10T19-33-48.638_00_signal-2022-08-10-153348.jpeg](./media/2022-08-10T19-33-48.638_00_signal-2022-08-10-153348.jpeg)  

[2022-08-10 19:34:11] Me:   [2022-08-10T19-34-11.986_00_Voice_Message_10-08-2022_15-34.m4a](./media/2022-08-10T19-34-11.986_00_Voice_Message_10-08-2022_15-34.m4a)  

"""  # noqa

expected_html = """<!DOCTYPE html>
<html lang="en">
    <head>
        <meta charset="utf-8"/>
        <title>
            Test
        </title>
        <link href="../style.css" rel="stylesheet"/>
    </head>
    <body>
        <div class="first">
            <a href="#pg0">
                FIRST
            </a>
        </div>
        <div class="last">
            <a href="#pg0">
                LAST
            </a>
        </div>
        <div class="page" id="pg0">
            <nav>
                <div class="prev">
                    PREV
                </div>
                <div class="next">
                    NEXT
                </div>
            </nav>
            <div class="msg me">
                <span class="date">
                    2022-08-10
                </span>
                <span class="time">
                    19:33:38
                </span>
                <span class="sender">
                    Me
                </span>
                <span class="body">
                    <p>
                        Test message
                    </p>
                </span>
                <span class="reaction">
                </span>
            </div>
            <div class="msg me">
                <span class="date">
                    2022-08-10
                </span>
                <span class="time">
                    19:33:48
                </span>
                <span class="sender">
                    Me
                </span>
                <span class="body">
                    <p>
                        Test image
                    </p>
                    <figure>
                        <label for="2022-08-10T19-33-48.638_00_signal-2022-08-10-153348.jpeg">
                            <img alt="2022-08-10T19-33-48.638_00_signal-2022-08-10-153348.jpeg" load="lazy" src="./media/2022-08-10T19-33-48.638_00_signal-2022-08-10-153348.jpeg"/>
                        </label>
                        <input class="modal-state" id="2022-08-10T19-33-48.638_00_signal-2022-08-10-153348.jpeg" type="checkbox"/>
                        <div class="modal">
                            <label for="2022-08-10T19-33-48.638_00_signal-2022-08-10-153348.jpeg">
                                <div class="modal-content">
                                    <img alt="2022-08-10T19-33-48.638_00_signal-2022-08-10-153348.jpeg" class="modal-photo" loading="lazy" src="./media/2022-08-10T19-33-48.638_00_signal-2022-08-10-153348.jpeg"/>
                                </div>
                            </label>
                        </div>
                    </figure>
                </span>
                <span class="reaction">
                </span>
            </div>
            <div class="msg me">
                <span class="date">
                    2022-08-10
                </span>
                <span class="time">
                    19:34:11
                </span>
                <span class="sender">
                    Me
                </span>
                <span class="body">
                    <audio controls="">
                        <source src="./media/2022-08-10T19-34-11.986_00_Voice_Message_10-08-2022_15-34.m4a" type="audio/mp4"/>
                    </audio>
                </span>
                <span class="reaction">
                </span>
            </div>
            <script>
                if (!document.location.hash) document.location.hash = 'pg0'
            </script>
        </div>
    </body>
</html>

"""  # noqa


def test_integration():
    from sigexport.main import main

    root = Path(__file__).resolve().parents[0]
    dest = Path(tempfile.TemporaryDirectory().name)
    source = root / "data"

    main(
        dest=dest,
        source=source,
        old=None,
        paginate=100,
        chats="",
        list_chats=False,
        include_empty=False,
        verbose=True,
        use_docker=False,
        docker_image="",
        print_data=False,
    )

    output_test = dest / "Test"
    output_media = output_test / "media"
    output_md = (output_test / "chat.md").read_text()
    output_html = (output_test / "index.html").read_text()

    print()
    print()
    print(output_md)
    print()
    print()

    assert expected_md == output_md
    assert expected_html == output_html

    assert (
        output_media / "2022-08-10T19-33-48.638_00_signal-2022-08-10-153348.jpeg"
    ).is_file()
    assert (
        output_media / "2022-08-10T19-34-11.986_00_Voice_Message_10-08-2022_15-34.m4a"
    ).is_file()

    shutil.rmtree(dest)



import os
import re
import shutil
from pathlib import Path

import markdown
from bs4 import BeautifulSoup
from typer import secho

from sigexport import models, templates
from sigexport.logging import log


def prep_html(dest: Path) -> None:
    """Prepare CSS etc"""
    root = Path(__file__).resolve().parents[0]
    css_source = root / "style.css"
    css_dest = dest / "style.css"
    if os.path.isfile(css_source):
        shutil.copy2(css_source, css_dest)
    else:
        secho(
            f"Stylesheet ({css_source}) not found."
            f"You might want to install one manually at {css_dest}."
        )


def create_html(
    name: str, messages: list[models.Message], msgs_per_page: int = 100
) -> str:
    """Create HTML version from Markdown input."""

    log(f"\tDoing html for {name}")
    # touch first
    ht_content = ""
    last_page = int(len(messages) / msgs_per_page)

    page_num = 0
    for i, msg in enumerate(messages):
        if i % msgs_per_page == 0:
            nav = "\n"
            if i > 0:
                nav += "</div>"
            nav += f"<div class=page id=pg{page_num}>"
            nav += "<nav>"
            nav += "<div class=prev>"
            if page_num != 0:
                nav += f"<a href=#pg{page_num-1}>PREV</a>"
            else:
                nav += "PREV"
            nav += "</div><div class=next>"
            if page_num != last_page:
                nav += f"<a href=#pg{page_num+1}>NEXT</a>"
            else:
                nav += "NEXT"
            nav += "</div></nav>\n"
            ht_content += nav
            page_num += 1

        sender = msg.sender
        date = msg.date.date().isoformat()
        time = msg.date.time().replace(microsecond=0).isoformat()

        reactions = " ".join(f"{r.name}: {r.emoji}" for r in msg.reactions)
        quote = ""
        if msg.quote:
            quote = f"<div class=quote>{msg.quote.replace('>', '')}</div>"

        body = msg.body
        try:
            body = markdown.Markdown().convert(body)
        except RecursionError:
            log(f"Maximum recursion on message {body}, not converted")

        # links
        p = re.compile(r"(https{0,1}://\S*)")
        a_template = r"<a href='\1' target='_blank'>\1</a> "
        body = re.sub(p, a_template, body)

        soup = BeautifulSoup(body, "html.parser")
        # attachments
        for att in msg.attachments:
            path = att.path
            src = f"./{path}"
            if models.is_image(path):
                temp = templates.figure.format(src=src, alt=att.name)
            elif models.is_audio(path):
                temp = templates.audio.format(src=src)
            elif models.is_video(path):
                temp = templates.video.format(src=src)
            else:
                temp = None
            if temp:
                soup.append(BeautifulSoup(temp, "html.parser"))

        cl = "msg me" if sender == "Me" else "msg"
        ht_content += templates.message.format(
            cl=cl,
            date=date,
            time=time,
            sender=sender,
            quote=quote,
            body=soup,
            reactions=reactions,
        )
    ht_text = templates.html.format(
        name=name,
        last_page=last_page,
        content=ht_content,
    )
    ht_text = BeautifulSoup(ht_text, "html.parser").prettify()
    ht_text = re.compile(r"^(\s*)", re.MULTILINE).sub(r"\1\1\1\1", ht_text)
    return ht_text



"""Main script for sigexport."""

import json
import shutil
import subprocess
import sys
from dataclasses import asdict
from pathlib import Path
from typing import Optional

from typer import Argument, Exit, Option, colors, run, secho

from sigexport import create, files, html, logging, merge, models, utils
from sigexport.logging import log

DATA_DELIM = "-----DATA-----"


# these are here because tiangolo/typer doesn't like Foo | None syntax
OptionalPath = Optional[Path]
OptionalStr = Optional[str]


def main(
    dest: Path = Argument(None),
    source: OptionalPath = Option(None, help="Path to Signal source database"),
    old: OptionalPath = Option(None, help="Path to previous export to merge"),
    paginate: int = Option(
        100, "--paginate", "-p", help="Messages per page in HTML; set to 0 for infinite"
    ),
    chats: str = Option(
        "", help="Comma-separated chat names to include: contact names or group names"
    ),
    json_output: bool = Option(
        False, "--json/--no-json", "-j", help="Whether to create JSON output"
    ),
    html_output: bool = Option(
        False, "--html/--no-html", "-h", help="Whether to create HTML output"
    ),
    list_chats: bool = Option(
        False, "--list-chats", "-l", help="List available chats and exit"
    ),
    include_empty: bool = Option(
        False, "--include-empty", help="Whether to include empty chats"
    ),
    overwrite: bool = Option(
        False,
        "--overwrite/--no-overwrite",
        help="Overwrite contents of output directory if it exists",
    ),
    verbose: bool = Option(False, "--verbose", "-v"),
    use_docker: bool = Option(
        True, help="Use Docker container for SQLCipher extraction"
    ),
    docker_image: str = Option(None, help="Docker image to use"),
    print_data: bool = Option(
        False, help="Print extracted DB data and exit (for use by Docker container)"
    ),
    _: bool = Option(False, "--version", callback=utils.version_callback),
) -> None:
    """Read the Signal directory and output attachments and chat to DEST directory."""
    logging.verbose = verbose

    if not any((dest, list_chats, print_data)):
        secho("Error: Missing argument 'DEST'", fg=colors.RED)
        raise Exit(code=1)

    if source:
        src = Path(source).expanduser().absolute()
    else:
        src = utils.source_location()
    source = src / "config.json"
    db_file = src / "sql" / "db.sqlite"

    # Read sqlcipher key from Signal config file
    if source.is_file():
        with open(source, encoding="utf-8") as conf:
            key = json.loads(conf.read())["key"]
    else:
        secho(f"Error: {source} not found in directory {src}")
        raise Exit(code=1)

    log(f"Fetching data from {db_file}\n")

    if use_docker:
        if not docker_image:
            docker_version = utils.VERSION.split(".dev")[0]
            docker_image = f"carderne/sigexport:v{docker_version}"
        secho(
            "Using Docker to extract data, this may take a while the first time!",
            fg=colors.BLUE,
        )
        cmd = [
            "docker",
            "run",
            "--rm",
            f"--volume={src}:/Signal",
            docker_image,
            "--no-use-docker",
            "--print-data",
        ]
        if chats:
            cmd.append(f"--chats={chats}")
        if include_empty:
            cmd.append("--include-empty")
        if verbose:
            cmd.append("--verbose")
        try:
            p = subprocess.run(
                cmd,  # NoQA: S603
                capture_output=True,
                text=True,
                check=False,
                encoding="utf-8",
            )
        except FileNotFoundError:
            secho("Error: using Docker method, but is Docker installed?", fg=colors.RED)
            secho("Try running this from the command line:\ndocker run hello-world")
            raise Exit(1)
        except subprocess.TimeoutExpired:
            secho("Docker process timed out.")
            raise Exit(1)
        try:
            docker_logs_1, data_raw, docker_logs_2 = p.stdout.split(DATA_DELIM)
        except ValueError:
            secho(f"Docker process failed, see logs below:\n{p.stderr}", fg=colors.RED)
            if sys.platform == "win32":
                secho("If the Signal app is still running, exit it and try again.")
            raise Exit(1)
        try:
            data = json.loads(data_raw)
            log(docker_logs_1)
            log(docker_logs_2)
        except json.JSONDecodeError:
            secho("Unable to decode data from Docker, see logs below:", fg=colors.RED)
            secho(p.stdout)
            secho(p.stderr, fg=colors.RED)
            raise Exit(1)
        try:
            convos_dict = data["convos"]
            contacts_dict = data["contacts"]

            convos: models.Convos = {}
            for k, cs in convos_dict.items():
                convos[k] = [models.RawMessage(**c) for c in cs]

            contacts: models.Contacts = {}
            for k, v in contacts_dict.items():
                contacts[k] = models.Contact(**v)

        except (KeyError, TypeError):
            secho(
                "Unable to extract convos and contacts from Docker, see data below",
                fg=colors.RED,
            )
            secho(data)
            raise Exit(1)
    else:
        try:
            from sigexport.data import fetch_data

            convos, contacts = fetch_data(
                db_file,
                key,
                chats=chats,
                include_empty=include_empty,
            )

            if print_data:
                convos_dict = {k: [asdict(c) for c in cs] for k, cs in convos.items()}
                contacts_dict = {k: asdict(v) for k, v in contacts.items()}
                data = {"convos": convos_dict, "contacts": contacts_dict}
                print(DATA_DELIM, json.dumps(data), DATA_DELIM)
                raise Exit()
        except (ImportError, ModuleNotFoundError):
            secho("You set 'no-use-docker' but `pysqlcipher3` not installed properly")
            sys.exit(1)

    if list_chats:
        names = sorted(v.name for v in contacts.values() if v.name is not None)
        secho(" | ".join(names))
        raise Exit()

    dest = Path(dest).expanduser()
    if not dest.is_dir():
        dest.mkdir(parents=True, exist_ok=True)
    elif not overwrite:
        secho(
            f"Output folder '{dest}' already exists, didn't do anything!", fg=colors.RED
        )
        raise Exit()

    contacts = utils.fix_names(contacts)

    secho("Copying and renaming attachments")
    for att_src, att_dst in files.copy_attachments(src, dest, convos, contacts):
        try:
            shutil.copy2(att_src, att_dst)
        except FileNotFoundError:
            secho(f"No file to copy at {att_src}, skipping!", fg=colors.MAGENTA)
        except OSError as exc:
            secho(f"Error copying file {att_src}, skipping!\n{exc}", fg=colors.MAGENTA)

    if json_output and old:
        secho(
            "Warning: currently, JSON does not support merging with the --old flag",
            fg=colors.RED,
        )

    secho("Creating output files")
    chat_dict = create.create_chats(convos, contacts)

    if old:
        secho(f"Merging old at {old} into output directory")
        secho("No existing files will be deleted or overwritten!")
        chat_dict = merge.merge_with_old(chat_dict, contacts, dest, Path(old))

    if paginate <= 0:
        paginate = int(1e20)

    if html_output:
        html.prep_html(dest)
    for key, messages in chat_dict.items():
        name = contacts[key].name
        # some contact names are None
        if not name:
            name = "None"
        md_path = dest / name / "chat.md"
        js_path = dest / name / "data.json"
        ht_path = dest / name / "index.html"

        md_f = md_path.open("a", encoding="utf-8")
        js_f = js_path.open("a", encoding="utf-8")
        ht_f = None
        if html_output:
            ht_f = ht_path.open("w", encoding="utf-8")

        try:
            for msg in messages:
                print(msg.to_md(), file=md_f)
                print(msg.dict_str(), file=js_f)
            if html_output:
                ht = html.create_html(
                    name=name, messages=messages, msgs_per_page=paginate
                )
                print(ht, file=ht_f)
        finally:
            md_f.close()
            js_f.close()
            if ht_f:
                ht_f.close()

    secho("Done!", fg=colors.GREEN)


def cli() -> None:
    """cli."""
    run(main)


if __name__ == "__main__":
    cli()



from pathlib import Path

from sigexport import models, utils
from sigexport.logging import log


def create_message(
    msg: models.RawMessage,
    name: str,  # only used for debug logging
    is_group: bool,
    contacts: models.Contacts,
) -> models.Message:
    ts = msg.get_ts()
    date = utils.dt_from_ts(ts)
    if ts == 0:
        log("\t\tNo timestamp or sent_at; date set to 1970")
    log(f"\t\tDoing {name}, msg: {date}")

    if msg.type == "call-history":
        body = (
            "Incoming call"
            if msg.call_history and msg.call_history["wasIncoming"]
            else "Outgoing call"
        )
    else:
        body = msg.body or ""

    body = body.replace("`", "")  # stop md code sections forming
    body += "  "  # so that markdown newlines

    sender = "No-Sender"
    if msg.type == "outgoing":
        sender = "Me"
    else:
        try:
            if is_group:
                for c in contacts.values():
                    num = c.number
                    if num is not None and num == msg.source:
                        sender = c.name
            else:
                sender = contacts[msg.conversation_id].name
        except KeyError:
            log(f"\t\tNo sender:\t\t{date}")

    attachments: list[models.Attachment] = []
    for att in msg.attachments:
        file_name = att["fileName"]
        path = Path("media") / file_name
        path = Path(str(path).replace(" ", "%20"))
        attachments.append(models.Attachment(name=file_name, path=str(path)))

    reactions: list[models.Reaction] = []
    if msg.reactions:
        for r in msg.reactions:
            try:
                reactions.append(
                    models.Reaction(contacts[r["fromId"]].name, r["emoji"])
                )
            except KeyError:
                log(
                    f"\t\tReaction fromId not found in contacts: "
                    f"[{date}] {sender}: {r}"
                )

    sticker = ""
    if msg.sticker:
        try:
            sticker = msg.sticker["data"]["emoji"]
        except KeyError:
            pass

    quote = ""
    if msg.quote:
        try:
            quote = msg.quote["text"].rstrip("\n")
            quote = quote.replace("\n", "\n> ")
            quote = f"\n\n> {quote}\n\n"
        except (AttributeError, KeyError, TypeError):
            pass

    return models.Message(
        date=date,
        sender=sender,
        body=body,
        quote=quote,
        sticker=sticker,
        reactions=reactions,
        attachments=attachments,
    )


def create_chats(
    convos: models.Convos,
    contacts: models.Contacts,
) -> models.Chats:
    """Convert convos and contacts into messages"""
    res: models.Chats = {}
    for key, raw_messages in convos.items():
        name = contacts[key].name
        log(f"\tDoing markdown for: {name}")
        is_group = contacts[key].is_group
        # some contact names are None
        if not name:
            name = "None"

        res[key] = [
            create_message(raw, name, is_group, contacts) for raw in raw_messages
        ]

    return res


# def create_member_lists(contacts: models.Contacts) -> str:
#



"""Extract data from Signal DB."""

import json
from pathlib import Path

from pysqlcipher3 import dbapi2 as sqlcipher

from sigexport import models
from sigexport.logging import log


def fetch_data(
    db_file: Path,
    key: str,
    chats: str,
    include_empty: bool,
) -> tuple[models.Convos, models.Contacts]:
    """Load SQLite data into dicts."""
    contacts: models.Contacts = {}
    convos: models.Convos = {}
    chats_list = chats.split(",") if len(chats) > 0 else []

    db = sqlcipher.connect(str(db_file))  # type: ignore
    c = db.cursor()
    # param binding doesn't work for pragmas, so use a direct string concat
    c.execute(f"PRAGMA KEY = \"x'{key}'\"")
    c.execute("PRAGMA cipher_page_size = 4096")
    c.execute("PRAGMA kdf_iter = 64000")
    c.execute("PRAGMA cipher_hmac_algorithm = HMAC_SHA512")
    c.execute("PRAGMA cipher_kdf_algorithm = PBKDF2_HMAC_SHA512")

    query = "SELECT type, id, e164, name, profileName, members FROM conversations"
    c.execute(query)
    for result in c:
        log(f"\tLoading SQL results for: {result[3]}, aka {result[4]}")
        members = []
        if result[5]:
            members = result[5].split(" ")
        is_group = result[0] == "group"
        cid = result[1]
        contacts[cid] = models.Contact(
            id=cid,
            name=result[3],
            number=result[2],
            profile_name=result[4],
            members=members,
            is_group=is_group,
        )
        if contacts[cid].name is None:
            contacts[cid].name = contacts[cid].profile_name

        if not chats or (result[3] in chats_list or result[4] in chats_list):
            convos[cid] = []

    query = "SELECT json, conversationId FROM messages ORDER BY sent_at"
    c.execute(query)
    for result in c:
        res = json.loads(result[0])
        cid = result[1]
        if cid and cid in convos:
            if res.get("type") in ["keychange", "profile-change"]:
                continue
            con = models.RawMessage(
                conversation_id=res["conversationId"],
                id=res["id"],
                type=res.get("type"),
                body=res.get("body", ""),
                contact=res.get("contact"),
                source=res.get("source"),
                timestamp=res.get("timestamp"),
                sent_at=res.get("sent_at"),
                server_timestamp=res.get("serverTimestamp"),
                has_attachments=res.get("has_attachments", False),
                attachments=res.get("attachments", []),
                read_status=res.get("read_status"),
                seen_status=res.get("seen_status"),
                call_history=res.get("call_history"),
                reactions=res.get("reactions", []),
                sticker=res.get("sticker"),
                quote=res.get("quote"),
            )
            convos[cid].append(con)

    if not include_empty:
        convos = {key: val for key, val in convos.items() if len(val) > 0}

    return convos, contacts



from __future__ import annotations

import json
import re
from collections import namedtuple
from dataclasses import asdict, dataclass
from datetime import datetime
from typing import Any


@dataclass
class RawMessage:
    conversation_id: str
    id: str

    body: str
    type: str | None
    contact: str | None
    source: Any | None

    timestamp: int | None
    sent_at: int | None
    server_timestamp: int | None

    has_attachments: bool
    attachments: list[dict[str, str]]

    read_status: bool | None
    seen_status: bool | None

    call_history: dict[str, Any] | None
    reactions: list[dict[str, Any]]
    sticker: dict[str, Any] | None
    quote: dict[str, Any] | None

    def get_ts(self: RawMessage) -> int:
        if self.sent_at and self.server_timestamp:
            if self.server_timestamp < self.sent_at:
                return self.server_timestamp
            else:
                return self.sent_at
        elif self.sent_at:
            return self.sent_at
        elif self.timestamp:
            return self.timestamp
        return 0


@dataclass
class Contact:
    id: str
    name: str
    number: str
    profile_name: str
    is_group: bool
    members: list[str] | None


Contacts = dict[str, Contact]
Convos = dict[str, list[RawMessage]]

Reaction = namedtuple("Reaction", ["name", "emoji"])


def is_image(p: str) -> bool:
    suffix = p.split(".")
    return len(suffix) > 1 and suffix[-1] in [
        "png",
        "jpg",
        "jpeg",
        "gif",
        "tif",
        "tiff",
    ]


def is_audio(p: str) -> bool:
    suffix = p.split(".")
    return len(suffix) > 1 and suffix[-1] in [
        "m4a",
        "aac",
    ]


def is_video(p: str) -> bool:
    suffix = p.split(".")
    return len(suffix) > 1 and suffix[-1] in [
        "mp4",
    ]


@dataclass
class Attachment:
    name: str
    path: str


@dataclass
class Message:
    date: datetime
    sender: str
    body: str
    quote: str
    sticker: str
    reactions: list[Reaction]
    attachments: list[Attachment]

    def to_md(self: Message) -> str:
        date_str = self.date.strftime("%Y-%m-%d %H:%M:%S")
        body = self.body

        if len(self.reactions) > 0:
            reactions = [f"{r.name}: {r.emoji}" for r in self.reactions]
            body = body + "\n(- " + ", ".join(reactions) + " -)"

        if len(self.sticker) > 0:
            body = body + "\n(( " + self.sticker + " ))"

        for att in self.attachments:
            if is_image(att.path):
                body += "!"
            body += f"[{att.name}](./{att.path})  "

        return f"[{date_str}] {self.sender}: {self.quote}{body}\n"

    def comp(self: Message) -> tuple[datetime, str, str]:
        date = self.date.replace(second=0, microsecond=0)
        return (date, self.sender, self.body.replace("\n", "").replace(">", ""))

    def dict(self: Message) -> dict:
        msg_dict = asdict(self)
        msg_dict["date"] = msg_dict["date"].isoformat()
        return msg_dict

    def dict_str(self: Message) -> str:
        return json.dumps(self.dict(), ensure_ascii=False)


Chats = dict[str, list[Message]]


@dataclass
class MergeMessage:
    date: datetime
    sender: str
    body: str

    def to_message(self: MergeMessage) -> Message:
        body = self.body

        p_reactions = re.compile(r"\n\(- (.*) -\)")
        m_reactions = re.findall(p_reactions, body)
        reactions = []
        if m_reactions:
            reactions = [Reaction(*r.split(":")) for r in m_reactions[0].split(", ")]
        body = re.sub(p_reactions, "", body)

        p_stickers = r"\n\(\( (.*) \)\)"
        stickers = re.findall(p_stickers, self.body)
        sticker = stickers[0] if stickers else ""
        body = re.sub(p_stickers, "", body)

        p_quote = re.compile(r"\n> (.*)$")
        m_quote = re.findall(p_quote, self.body)
        quote = ""
        if m_quote:
            quote = "".join(m_quote)
        body = re.sub(p_quote, "", body)

        p_attachments = r"!{0,1}\[(.*?)\]\((.*?)\)"
        m_attachments = re.findall(p_attachments, self.body)
        attachments = []
        if m_attachments:
            attachments = [Attachment(name=g[0], path=g[1]) for g in m_attachments]
        body = re.sub(p_attachments, "", body)

        body = body.rstrip("\n")

        return Message(
            date=self.date,
            sender=self.sender,
            body=body,
            quote=quote,
            reactions=reactions,
            sticker=sticker,
            attachments=attachments,
        )



import re
import shutil
from pathlib import Path

from sigexport import files, models, utils
from sigexport.logging import log


def lines_to_msgs(lines: list[str]) -> list[models.MergeMessage]:
    """Extract messages from lines of Markdown."""
    p = re.compile(r"^\[(\d{4}-\d{2}-\d{2},? \d{2}:\d{2}(?::\d{2})?)] (.*?): ?(.*\n)")
    msgs: list[models.MergeMessage] = []
    for li in lines:
        m = p.match(li)
        if m:
            date_str, sender, body = m.groups()
            date = utils.parse_datetime(date_str)
            msg = models.MergeMessage(date=date, sender=sender, body=body)
            msgs.append(msg)
        else:
            msgs[-1].body += li
    return msgs


def merge_chat(new: list[models.Message], path_old: Path) -> list[models.Message]:
    """Merge new and old chat markdowns."""
    with path_old.open(encoding="utf-8") as f:
        old_raw = f.readlines()

    old = lines_to_msgs(old_raw)
    old_msgs = [o.to_message() for o in old]

    try:
        a = old_raw[0][:30]
        b = old_raw[-1][:30]
        c = new[0].to_md()[:30]
        d = new[-1].to_md()[:30]
        log(f"\t\tFirst line old:\t{a}")
        log(f"\t\tLast line old:\t{b}")
        log(f"\t\tFirst line new:\t{c}")
        log(f"\t\tLast line new:\t{d}")
    except IndexError:
        log("\t\tNo new messages for this conversation")

    # get rid of duplicates
    msg_dict = {m.comp(): m for m in old_msgs + new}
    merged = list(msg_dict.values())

    return merged


def merge_with_old(
    chat_dict: models.Chats, contacts: models.Contacts, dest: Path, old: Path
) -> models.Chats:
    """Main function for merging new and old."""
    new_chat_dict: models.Chats = {}
    for key, msgs in chat_dict.items():
        name = contacts[key].name
        # some contact names are None
        if not name:
            name = "None"
        dir_old = old / name
        if dir_old.is_dir():
            log(f"\tMerging {name}")
            dir_new = dest / name
            if dir_new.is_dir():
                files.merge_attachments(dir_new / "media", dir_old / "media")
                try:
                    path_old = dir_old / "chat.md"
                    msgs_new = merge_chat(msgs, path_old)
                    new_chat_dict[key] = msgs_new
                except FileNotFoundError:
                    try:
                        path_old = dir_old / "index.md"  # old name
                        msgs_new = merge_chat(msgs, path_old)
                        new_chat_dict[key] = msgs_new
                    except FileNotFoundError:
                        log(f"\tNo old for {name}")
            else:
                shutil.copytree(dir_old, dir_new)
    return new_chat_dict
